
#pragma Ospace

#include <nds.h>

#include "_console.h"
#include "_const.h"
#include "maindef.h"
#include "unicode.h"
#include "strtool.h"
#include "procstate.h"
#include "launchstate.h"
#include "shell.h"
#include "lang.h"

#include "BootROM.h"

TBootROMInfo BootROMInfo;

// ------------------------------------

void BootROM_Init(void)
{
  BootROMInfo.PathUnicode[0]=0;
  BootROMInfo.FilenameUnicode[0]=0;
  BootROMInfo.FullPathAlias[0]=0;
  BootROMInfo.isNDSHomeBrew=false;
  BootROMInfo.RequestBackupSave=false;
  BootROMInfo.Execute=false;
}

bool BootROM_GetExecuteFlag(void)
{
  return(BootROMInfo.Execute);
}

const char* BootROM_GetFullPathAlias(void)
{
  return(BootROMInfo.FullPathAlias);
}

// ------------------------------------

void BootROM_SetInfo(const UnicodeChar *pPathUnicode,const UnicodeChar *pFilenameUnicode,bool RequestBackupSave)
{
  const UnicodeChar *pFullPathUnicode=ConvertFull_MargeFromSplit(pPathUnicode,pFilenameUnicode);
  LaunchState_Add(ELST_NDS,pFullPathUnicode);
  LaunchState.LastTab=ELST_NDS;
  LaunchState_Save();
      
  if(ProcState.System.AutoLastState==true){
    ProcState.System.LastState=ELS_Launch;
    ProcState_RequestSave=true;
  }
  
  const char *palias=ConvertFull_Unicode2Alias(pPathUnicode,pFilenameUnicode);
  
  if((palias==NULL)||(palias[0]==0)){
    _consolePrintf("Fatal error: Can not convert unicode to alias.\n");
    ShowLogHalt();
  }
  
  BootROMInfo.RequestBackupSave=RequestBackupSave;
  
  Unicode_Copy(BootROMInfo.PathUnicode,pPathUnicode);
  Unicode_Copy(BootROMInfo.FilenameUnicode,pFilenameUnicode);
  StrCopy(palias,BootROMInfo.FullPathAlias);
  BootROMInfo.Execute=false;
  
  SetNextProc(ENP_BootROM,EPFE_None);
}

void BootROM_SetInfo_NoLaunch(const UnicodeChar *pPathUnicode,const UnicodeChar *pFilenameUnicode,bool RequestBackupSave)
{
  const char *palias=ConvertFull_Unicode2Alias(pPathUnicode,pFilenameUnicode);
  
  if((palias==NULL)||(palias[0]==0)){
    _consolePrintf("Fatal error: Can not convert unicode to alias.\n");
    ShowLogHalt();
  }
  
  BootROMInfo.RequestBackupSave=RequestBackupSave;
  
  Unicode_Copy(BootROMInfo.PathUnicode,pPathUnicode);
  Unicode_Copy(BootROMInfo.FilenameUnicode,pFilenameUnicode);
  StrCopy(palias,BootROMInfo.FullPathAlias);
  BootROMInfo.Execute=false;
  
  SetNextProc(ENP_BootROM,EPFE_None);
}

// ------------------------------------

void BootROM_Execute(void)
{
  u8 header[16];
  {
    FAT_FILE *pf=Shell_FAT_fopen_Split(BootROMInfo.PathUnicode,BootROMInfo.FilenameUnicode);
    if(pf==NULL){
      _consolePrint("Can not open NDS file.\n");
      ShowLogHalt();
    }
    FAT2_fread(header,1,16,pf);
    FAT2_fclose(pf);
  }
  
  char ID[5];
  ID[0]=header[12+0];
  ID[1]=header[12+1];
  ID[2]=header[12+2];
  ID[3]=header[12+3];
  ID[4]=0;
  
  _consolePrintf("Detected ROMID: %s\n",ID);
  
  bool homebrew=false;
  
  if(strcmp("####",ID)==0) homebrew=true;
  if(strcmp("PASS",ID)==0) homebrew=true;
  if((ID[0]==0x3d)&&(ID[1]==0x84)&&(ID[2]==0x82)&&(ID[3]==0x0a)) homebrew=true;
  
  if(homebrew==false){
    _consolePrintf("Fatal error: Not support commercial NDSROM file.\n");
    ShowLogHalt();
  }
  
  BootROMInfo.isNDSHomeBrew=true;
  BootROMInfo.Execute=true;
}

bool BootROM_isHomeBrew(void)
{
  return(BootROMInfo.isNDSHomeBrew);
}

